﻿
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace exc.jdbi.Pointers;

public sealed class GcIPtr<T> : IDisposable where T : struct
{
  public IntPtr Ptr
  {
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    get => (IntPtr)this;
  }

  public int TypeSize { get; private set; }
  public bool IsDisposed { get; private set; }
  internal GCHandle PinnedArray { get; private set; }
  public T[] Target
  {
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    get
    {
      if (this.PinnedArray.Target is not null)
        return (T[])this.PinnedArray.Target;

      throw new ArgumentNullException(nameof(this.PinnedArray));
    }
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe GcIPtr()
  {
    if (!typeof(T).IsPrimitive)
      throw new NotImplementedException(nameof(GcIPtr<T>));

    this.TypeSize = Marshal.SizeOf(typeof(T));
    this.TypeSize = typeof(T) == typeof(char) ? 2 : this.TypeSize;
    this.TypeSize = typeof(T) == typeof(bool) ? 1 : this.TypeSize;
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public GcIPtr(T[] pinned_array) : this()
  {
    this.NewInit(pinned_array);
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe void NewInit(T[] pinned_array)
  {
    this.AssertGcIPtr(true);

    this.Dispose();
    this.IsDisposed = false;
    this.PinnedArray = GCHandle.Alloc(pinned_array, GCHandleType.Pinned);
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe T[] ToValues()
  {
    this.AssertGcIPtr(false);

    if (this.PinnedArray.Target is not null)
      return ((T[])this.PinnedArray.Target).ToArray();
    return Array.Empty<T>();
  }


  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe byte[] ToArray()
  {
    this.AssertGcIPtr(false);

    if (this.PinnedArray.Target is not null)
    {
      if (typeof(T) == typeof(byte))
        return (byte[])this.PinnedArray.Target;

      Span<T> result;
      var values = (T[])this.PinnedArray.Target;
      var bytes = new byte[values.Length * this.TypeSize];
      fixed (byte* ptr = bytes)
        result = new Span<T>(ptr, values.Length);
      values.AsSpan().CopyTo(result);
      return bytes;
    }

    throw new ArgumentNullException(nameof(this.PinnedArray));
  }

  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static implicit operator IntPtr(GcIPtr<T> value)
  {
    if (value.IsDisposed)
      throw new ObjectDisposedException(nameof(GcIPtr<T>));

    //Weak-Pointer?!
    return value.PinnedArray.AddrOfPinnedObject();
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public void Dispose()
  {
    if (this.IsDisposed) return;

    this.IsDisposed = true;
    if (this.PinnedArray.IsAllocated)
      this.PinnedArray.Free();

    //!!!!!!!!!!! !!!!!!!!!!! !!!!!!!!!!! !!!!!!!!!!! 
    //IsAllocated(??) sollte eigentlich auf false stehen.
    //!!!!!!!!!!! !!!!!!!!!!! !!!!!!!!!!! !!!!!!!!!!! 
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  private void AssertGcIPtr(bool _onlydispose)
  {
    if (this.IsDisposed)
      throw new ObjectDisposedException(nameof(GcIPtr<T>));

    if (_onlydispose) return;

    if (this.PinnedArray.Target is null)
      throw new ArgumentNullException(
        null, nameof(this.PinnedArray));
  }
}
