﻿
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace exc.jdbi.Pointers;

public sealed class UsIPtr<T> : IDisposable where T : struct
{
  public IntPtr Ptr { get; private set; }
  public int TypeSize { get; private set; }
  public int Length { get; private set; }
  public bool IsDisposed { get; private set; }

  public unsafe Span<T> Target
  {
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    get => new(this.Ptr.ToPointer(), this.Length);
  }

  public unsafe T this[int index]
  {
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    get => new Span<T>(this.Ptr.ToPointer(), this.Length)[index];
    [MethodImpl(MethodImplOptions.AggressiveInlining)]
    set => new Span<T>(this.Ptr.ToPointer(), this.Length)[index] = value;
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public UsIPtr()
  {
    if (!typeof(T).IsPrimitive)
      throw new NotImplementedException(nameof(GcIPtr<T>));

    this.TypeSize = Marshal.SizeOf(typeof(T));
    this.TypeSize = typeof(T) == typeof(char) ? 2 : this.TypeSize;
    this.TypeSize = typeof(T) == typeof(bool) ? 1 : this.TypeSize;
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public UsIPtr(T[] values) : this()
    => this.NewInit(values);

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe void NewInit(T[] values)
  {
    if (this.IsDisposed)
      throw new ObjectDisposedException(nameof(UsIPtr<T>));

    this.Dispose();
    this.IsDisposed = false;
    this.Length = values.Length;
    //alloc unmanaged memory
    this.Ptr = this.ToIntPtr(values);
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe byte[] ToBytes()
    => new Span<byte>(this.Ptr.ToPointer(), this.TypeSize * this.Length).ToArray();

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe Span<byte> AsBytes()
      => new(this.Ptr.ToPointer(), this.TypeSize * this.Length);

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe T[] ToValues()
    => new Span<T>(this.Ptr.ToPointer(), this.Length).ToArray();

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static implicit operator IntPtr(UsIPtr<T> value)
    => value.Ptr;

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public unsafe void Dispose()
  {
    if (this.IsDisposed) return;

    this.IsDisposed = true;
    var bytes = (byte*)this.Ptr;
    var length = this.Length * this.TypeSize;
    for (var i = 0; i < length; i++)
      *(bytes + i) = 0;
    Marshal.FreeHGlobal(this.Ptr);
    this.Ptr = IntPtr.Zero;
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  private unsafe IntPtr ToIntPtr(T[] values)
  {
    var result = Marshal.AllocHGlobal(this.Length * this.TypeSize);
    var span = new Span<T>(result.ToPointer(), values.Length);
    values.AsSpan().CopyTo(span);
    return result;
  }
}


