﻿
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace exc.jdbi.Spans;

//Wichtig: Der Cast von einem Typ in den anderen, kann unter Umständen 
//         problematisch sein, da der gleiche Speicherbereich genutzt 
//         wird, und somit die Byte-Zahl nicht übereinstimmt muss. 
//Beispiel: Int16 zu Int32, wenn Int16 size = 51 ist.
//          Da 51 * 2 mod 4 != 0 ist

//SpanTool besitzt bei Cast die option 'cut_admit'
//Der Basiswert ist false, und gibt im oberen Beispiel ein Exception aus.
//Sofern true wird einfach abgeschnitten! >> 51 * 2 div 4 = 25,
//da nur der jeweilige Speicherbereich genutzt wird.

//******** ******** ******** ******** ******** ******** ******** 
//Sofern das nicht erwünscht ist die Copy-Methoden nutzen.
//******** ******** ******** ******** ******** ******** ******** 



//Important: The same memory area is used, which can lead to problems.
//Example:   Int16 to Int32 if Int16 size = 51.
//           Since 51 * 2 mod 4 != 0

//SpanTool has the option 'cut_admit' in Cast.
//The base value is 'false', and gives an exception in the example above.
//If 'true' it is simply cut off! >> 51 * 2 div 4 = 25,
//because only the respective memory area is used.

//******** ******** ******** ******** ******** ******** ******** 
//If this is not desired use the copy methods.
//******** ******** ******** ******** ******** ******** ******** 


partial class SpanTool
{

  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static ReadOnlySpan<TOUT> Cast<TIN, TOUT>(ReadOnlySpan<TIN> span, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    return Cast<TIN, TOUT>(span, 0, span.Length, cut_admit);
  }


  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static Span<TOUT> Cast<TIN, TOUT>(Span<TIN> span, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    return Cast<TIN, TOUT>(span, 0, span.Length, cut_admit);
  }

  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static ReadOnlySpan<TOUT> Cast<TIN, TOUT>(ReadOnlySpan<TIN> span, int start, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    return Cast<TIN, TOUT>(span, start, span.Length - start, cut_admit);
  }


  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static Span<TOUT> Cast<TIN, TOUT>(Span<TIN> span, int start, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    return Cast<TIN, TOUT>(span, start, span.Length - start, cut_admit);
  }

  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static ReadOnlySpan<TOUT> Cast<TIN, TOUT>(ReadOnlySpan<TIN> span, int start, int length, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    if (span.Length - start - length < 0)
      throw new ArgumentOutOfRangeException(nameof(length));

    var sztin = SizeOfT<TIN>();
    var sztout = SizeOfT<TOUT>();

    if (!cut_admit)
      if ((length * sztin % sztout) != 0)
        throw new ArgumentOutOfRangeException(nameof(length));
    return MemoryMarshal.Cast<TIN, TOUT>(span.Slice(start, length));
  }


  [SkipLocalsInit]
  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public static Span<TOUT> Cast<TIN, TOUT>(Span<TIN> span, int start, int length, bool cut_admit = false)
    where TIN : struct where TOUT : struct
  {
    //return the same memory
    if (span.Length - start - length < 0)
      throw new ArgumentOutOfRangeException(nameof(length));

    var sztin = SizeOfT<TIN>();
    var sztout = SizeOfT<TOUT>();
    if (!cut_admit)
      if ((length * sztin % sztout) != 0)
        throw new ArgumentOutOfRangeException(nameof(length));
    return MemoryMarshal.Cast<TIN, TOUT>(span.Slice(start, length));
  }
}