﻿

using System.Text;
using exc.jdbi.Spans; 
using System.Runtime.CompilerServices;
using exc.jdbi.Pointers;

namespace SpanToolTest;

using static RandomHolder;

internal class TestSpanTool
{

  public static void Start()
  {
    TestSizeOfT();
    TestAsArray();
    TestAsRef();
    TestCast();
    TestConvert();
    TestCopy();
    TestToString();
    TestToAsSpan();
    TestBinarySearch();
    TestClear();
    TestConcat();
    TestContains();
    TestCreateInstance();
    TestEmpty();
    TestFill();
    TestForEach();
    TestIndexOf();
    TestReverse();
    TestSlice();
    TestStrings();
    TestSwap();
    TestToUpperLower();
    TestPointers();
  }

  private static void TestSizeOfT()
  {
    var size = SpanTool.SizeOfT<uint>();
  }


  private static void TestAsArray()
  {
    // Wichtig: Der gleiche Speicherbereich wird zurückgegeben
    // Important: The same memory area Is returned.

    // Wichtig: Wird die Source verändert, ändert sich 
    // auch die Byte-Destination !

    // Es wird kein neuer Speicher alloziert.

    var bytes = RngBytes(Rand.Next(20, 100));
    var span1 = SpanTool.AsBytes<byte>(bytes);
    span1[0] = 1; // In bytes ändert es sich auch
    bytes[0] = 10; // genauso auch jetzt in span1

    var uints = RngInts<uint>(Rand.Next(20, 100));
    var span2 = SpanTool.AsBytes<uint>(uints);
    uints[0] = 1; //siehe in span2

  }

  private static void TestAsRef()
  {
    // Zurzeit ist in Vb.Net der Rückgabewert einer Funktion
    // als 'ByRef' nicht möglich. (über Parameterliste schon)

    //In C# hingegen kann mit 'ref' eine referenzierte
    //Deklaration gemacht werden, die im Ursprung auch ändert.


    var bytes = RngBytes(Rand.Next(5, 10)).AsSpan();
    ref byte spanbytesref = ref SpanTool.AsRef<byte>(bytes);
    spanbytesref = 1; // In bytes ändert sich es.

    // Der zurückgegebene Wert aus 'AsRef' wird in eine
    // neue Variable auf dem Stack geschrieben.

    var spanbytes = SpanTool.AsRef<byte>(bytes);
    spanbytes = 2; // In bytes ändert sich nichts.


    bytes = RngBytes(Rand.Next(5, 10)).AsSpan();
    ref short spani16ref = ref SpanTool.AsRef<short>(bytes);
    spani16ref = 1; // In bytes ändert sich es.

    var spani16 = SpanTool.AsRef<short>(bytes);
    spani16 = 2; // In bytes ändert sich nichts.

  }

  private static void TestCast()
  {
    // Wichtig: Der Cast von einem Typ in den anderen, kann unter Umständen 
    // problematisch sein, da der gleiche Speicherbereich genutzt 
    // wird, und somit die Byte-Zahl nicht übereinstimmt muss. 
    // Beispiel: Int16 zu Int32, wenn Int16 size = 51 ist.
    // Da 51 * 2 mod 4 != 0 ist

    // SpanTool besitzt bei Cast die option 'cut_admit'
    // Der Basiswert ist false, und gibt im oberen Beispiel ein Exception aus.
    // Sofern true wird, einfach abgeschnitten! >> 51 * 2 div 4 = 25,
    // da nur der jeweilige Speicherbereich genutzt werden kann.

    // ******** ******** ******** ******** ******** ******** ******** 
    // Sofern das nicht erwünscht ist die Copy-Methoden nutzen.
    // ******** ******** ******** ******** ******** ******** ******** 


    var uints = RngInts<uint>(Rand.Next(5, 10));
    var spanbytes = SpanTool.Cast<uint, byte>(uints);
    uints[0] = 1; // spanbytes ändert sich, da der gleiche Speicherbereich


    var i16s = RngInts<short>(51);
    // Exception wenn size = 51 und cut_admit = false
    var spanints = SpanTool.Cast<short, int>(i16s, true); // Wichtig: size = 25
    //spanints[0] = 1; // i16s würde sich ändert


    // Die Copy-Methode löst das Problem, wobei hier zwei verschiedene 
    // Speicherbereiche (hier beide im Heap) genutzt werden.
    var ints = new int[26]; // 51 * 2 / 4 --> 26
    SpanTool.Copy<short, int>(i16s, 0, ints, 0, i16s.Length);


    //spanints != ints

  }

  private static void TestConvert()
  {

    // Bei einer Konvertierung werden in einer Array mit
    // primitiven Datentypen, die einzelnen Werte nicht als
    // Gesamtwert (zusammenhängend) betrachtet, sondern
    // jedes einzelne Item (Zahl) wird in ein bestimmter
    // Datentyp konvertiert. 
    // >> Anders als bei Copy(..)


    var uints1 = RngInts<uint>(Rand.Next(5, 10));
    var spani16s = SpanTool.Convert<uint, short>(uints1);

    //Mit 'Array.ConvertAll' kann es so gemacht werden
    var converter = new Converter<uint, short>(ToT<uint, short>);
    var i16s = Array.ConvertAll(uints1, converter);
    var uints2 = Array.ConvertAll(i16s, new Converter<short, uint>(ToT<short, uint>));

    //uints1 == uints2 genauso auch spani16s == i16s


    //Hier noch eine Char-Variante
    var chars = RngInts<char>(Rand.Next(10, 20));
    var spanuints = SpanTool.Convert<char, uint>(chars);
  }


  private static void TestCopy()
  {
    //Copy eignet sich hervorragend, wenn verschiedene 
    //Speicherbereiche vorhanden sind (auch in der Länge). 
    //Es können Heap, Stack, Unmanaged Speicherbereiche verwendet werden.

    var i16s = RngInts<short>(Rand.Next(5, 10));
    var size = i16s.Length * SpanTool.SizeOfT<short>() % SpanTool.SizeOfT<int>() == 0
            ? i16s.Length * SpanTool.SizeOfT<short>() / SpanTool.SizeOfT<int>()
            : (i16s.Length * SpanTool.SizeOfT<short>() / SpanTool.SizeOfT<int>()) + 1;
    var ints = new int[size];
    SpanTool.Copy<short, int>(i16s, 0, ints, 0, i16s.Length);


    var uints1 = RngInts<uint>(Rand.Next(5, 10));
    var uints2 = new uint[uints1.Length];
    SpanTool.Copy<uint>(uints1, 0, uints2, 0, uints1.Length);

  }

  private static void TestToString()
  {
    //Macht, was gemacht werden muss

    var bytes = RngBytes(Rand.Next(5, 20));
    var str = SpanTool.ToString<byte>(bytes, " ");

    var uints = RngInts<uint>(Rand.Next(5, 20));
    str = SpanTool.ToString<uint>(uints, " ");

    var i16s = RngInts<short>(Rand.Next(5, 20));
    str = SpanTool.ToString<short>(i16s, " ");

  }

  private static void TestToAsSpan()
  {
    var number = RngInt<int>();

    //Der gleiche Speicherbereich wird zurückgegeben.
    var spanref = SpanTool.AsSpan(ref number);
    spanref[0] = 123456;//number ändert sich


    var span = SpanTool.ToNewSpan(number);
    span[0] = 654321;//number ändert sich

  }

  private static void TestBinarySearch()
  {
    //https://github.com/dotnet/runtime/issues/20107
    //https://www.geeksforgeeks.org/how-to-use-array-binarysearch-method-in-c-sharp-set-1/


    // Sucht den gewünschten Value in einer >sortierten< Span
    // >> Ähnlich wie Contains

    var i32value = RngInt<int>();
    var i32s = RngInts<int>(10);
    var i32ssort = i32s.ToArray();

    // Array muss sortiert sein
    Array.Sort(i32ssort);
    var index = i32ssort.AsSpan().BinarySearch(i32value); //Nicht gefunden

    var comparer = new InlineableComparable();
    var cntval = i32s.Count(x => x == i32value);
    index = SpanTool.BinarySearch(i32ssort, comparer); //Nicht gefunden


    i32value = i32s[Rand.Next(i32s.Length)];
    comparer = new InlineableComparable(i32value);
    cntval = i32s.Count(x => x == i32value);
    index = SpanTool.BinarySearch(i32ssort, comparer);  // gefunden
  }


  private static void TestClear()
  {
    var uints = RngInts<uint>(Rand.Next(5, 20));
    SpanTool.Clear<uint>(uints);
  }

  private static void TestConcat()
  {
    //Der neue Speicher für die Zusammensetzung
    //kommt immer aus dem Heap


    var bytes1 = RngBytes(Rand.Next(10, 30));
    var bytes2 = RngBytes(Rand.Next(10, 30));
    var concat = SpanTool.Concat<byte>(bytes1, bytes2);
  }


  private static void TestContains()
  {
    var i16value = RngInt<short>();
    var i16s = RngInts<short>(Rand.Next(10, 30));
    i16s[Rand.Next(i16s.Length)] = i16value;
    var result = SpanTool.Contains(i16s, i16value);
    //var index = SpanTool.IndexOf(i16s, i16value);

    var chs = RngAlphaString(Rand.Next(10, 30));
    var chvalue = chs[Rand.Next(chs.Length)];
    result = SpanTool.Contains(chs, chvalue);
    //index = SpanTool.IndexOf(chs, chvalue);

    result = SpanTool.Contains(chs, new[] { chvalue }, StringComparison.CurrentCulture);
    //index = SpanTool.IndexOf(chs, new[] { chvalue }, StringComparison.CurrentCulture);

    chs = RngAlphaString(Rand.Next(10, 30));
    var start = Rand.Next(chs.Length / 3);
    var length = Rand.Next(3, 10);
    var chvalues = chs.AsSpan().Slice(start, length).ToArray();
    result = SpanTool.Contains(chs, chvalues, StringComparison.CurrentCulture);
    //index = SpanTool.IndexOf(chs, chvalues, StringComparison.CurrentCulture);
  }

  private static void TestCreateInstance()
  {
    var spanuints = SpanTool.CreateInstance<uint>(32);
    spanuints = (uint[])SpanTool.CreateInstance(typeof(uint),32);

    var spanuintsjagged = SpanTool.CreateInstance<uint>(8, 16, 32, 64);

    //false for jagged
    var tmp = (Array[])SpanTool.CreateInstance(typeof(uint), new[] { 8, 16, 32, 64 }, false);
    spanuintsjagged = tmp.Select(x=>(uint[])x).ToArray();

  }

  private static void TestEmpty()
  {
    var span = SpanTool.Empty<uint>();

  }

  private static void TestFill()
  {
    var uints = new uint[10];
    SpanTool.Fill(uints,uint.MaxValue);
  }

  private static void TestForEach()
  {
    var i16s = RngInts<short>(Rand.Next(10, 30));
    SpanTool.ForEach<short>(i16s,x=> Console.Write($"{x} ")); 
    Console.WriteLine();

    SpanTool.ForEach<short>(i16s, (x, idx) => Console.Write((idx % 2) == 0 ? $"{x} " : string.Empty));
    Console.WriteLine();
  }

  private static void TestIndexOf()
  {
    var bvalue = RngInt<byte>();
    var bytes = RngBytes(Rand.Next(20,50));
    var result = SpanTool.IndexOf(bytes, bvalue);


    var i16value = RngInt<short>();
    var i16s = RngInts<short>(Rand.Next(20,50));
    i16s[Rand.Next(i16s.Length)] = i16value;
    result = SpanTool.IndexOf(i16s, i16value);


    var i32value = RngInts<int>(3);
    var i32s = Enumerable.Repeat(1, 100).ToArray();
    result = SpanTool.IndexOf<int>(i32s, i32value);
    i32value.AsSpan().CopyTo(i32s.AsSpan().Slice(Rand.Next(i32s.Length / 3), i32value.Length));
    result = SpanTool.IndexOf<int>(i32s, i32value);


    bvalue = RngInt<byte>();
    bytes = RngBytes(Rand.Next(128, 1000));
    var resultall = SpanTool.IndexOfAll(bytes, bvalue);
    //var rcntall = bytes.Count(x => x == bvalue);


    //und noch einige andere Möglichkeiten...

  }


  private static void TestReverse()
  {

    var bytes1 = RngBytes(Rand.Next(10,20)); 
    SpanTool.Reverse<byte>(bytes1);

    bytes1 = RngBytes(Rand.Next(10, 20));
    var start = Rand.Next(2, 5); 
    SpanTool.Reverse<byte>(bytes1, start);

    bytes1 = RngBytes(Rand.Next(10, 20));
    var length = Rand.Next(3, 6);
    SpanTool.Reverse<byte>(bytes1, start, length); 

  }


  private static void TestSlice()
  {
    var bytes = RngBytes(Rand.Next(10, 20));
    var slice1 = bytes.AsSpan().Slice( 2, 5);
    var slice2 = SpanTool.Slice<byte>(bytes, 2, 5);

    //Alloc memory from heap
    var skip = SpanTool.Skip<byte>(bytes, 2);
    var take = SpanTool.Take<byte>(bytes, 5);
    var skiptake = SpanTool.SkipTake<byte>(bytes, 2, 5);
  }


  private static void TestStrings()
  {
    string s = "exc-jdbi";
    var str = RngAlphaString(Rand.Next(128, 513));
    var strspan = str.ToCharArray().AsSpan();
    SpanTool.Insert(strspan, ' ', 0);
    SpanTool.Insert(strspan, s, 0);

    var newstr = SpanTool.Insert(str, ' ', 0);
    newstr = SpanTool.Insert(str, s, 0);

    str = new string('a', 10);
    newstr = SpanTool.Replace(str, 'b', 'a');
    newstr = SpanTool.Replace(str, 'a', 'b');

    str = "acacacacacacacacacac";
    newstr = SpanTool.Replace(str, 'b', 'a');
    newstr = SpanTool.Replace(str, 'a', 'b');
    newstr = SpanTool.Replace(str, 'a', s);
    newstr = SpanTool.Replace(str, "a", s);

    str = "akackackackackackackackackaaca";
    newstr = SpanTool.Replace(str, "ac", s, 0, 29);
  }

  private static void TestSwap()
  {
    var bytes = RngBytes(Rand.Next(10, 20));
    SpanTool.Swap<byte>(bytes,1,3);
  }


  private static void TestToUpperLower()
  {
    //new alloc memory from heap
    var chs = RngAlphaString(Rand.Next(5, 10));
    var lower = SpanTool.ToLower(chs); 
    var upper = SpanTool.ToUpper(chs);
  }

  private static void TestPointers()
  {

    // Die Zahlen in dieser Array werden verändert,
    // gleichzeitig ist die Array wie "fixiert"
    var uints = RngInts<uint>(4);

    using (var gcptr = new GcIPtr<uint>(uints))
    {
      var target = gcptr.Target;

      var copybytes = gcptr.ToArray();
      copybytes[1] = 3;

      var copyvalues = gcptr.ToValues();
      copyvalues[1] = 3;

      target[0] = 5;
      target[2] = 5;

      var ptr = gcptr.Ptr;

      var value = gcptr.Target[2];
      value = gcptr.Target[2];
    }

    // Die Array wird in den unmanaged Bereich
    // kopiert. Im Unmanaged ist die Array wie "fixiert".
    uints = RngInts<uint>(4);

    // UsIPtr bietet minimale Funktionen,
    // die sonst in der Praxis gebraucht werden.
    using (var usptr = new UsIPtr<uint>(uints))
    {
      var value = usptr.Target[2];

      var target = usptr.Target;
      target[2] = 5;
      usptr[0] = 5;

      var ptr = usptr.Ptr;

      var copybytes = usptr.ToBytes();
      copybytes[1] = 2;

      var copyvalues = usptr.ToValues();
      copyvalues[1] = 1;
    }


    // Und nochmals GcIPtr
    uints = RngInts<uint>(4);
    var gcptr2 = new GcIPtr<uint>(uints);
    var target2 = gcptr2.Target;
    var copybytes2 = gcptr2.ToArray();

    target2[0] = 5;
    target2[2] = 5;

    var ptr2 = gcptr2.Ptr;

    var value2 = gcptr2.Target[2];
    gcptr2.Dispose();

    // Und nochmals UsIPtr
    uints = RngInts<uint>(4);
    var usptr2 = new UsIPtr<uint>(uints);
    value2 = usptr2.Target[2];

    var target3 = usptr2.Target;
    target3[2] = 5;
    usptr2[0] = 5;

    ptr2 = usptr2.Ptr;

    copybytes2 = usptr2.ToBytes();
    var copyvalues2 = usptr2.ToValues();
    usptr2.Dispose();

  }

  private static TOUT ToT<TIN, TOUT>(TIN number)
    where TIN : struct, IConvertible, IComparable
    where TOUT : struct, IConvertible
  {
    return SpanTool.Convert<TIN, TOUT>(number);
  }

  private static string RngAlphaString(int size)
  {
    var alpha = AlphaString().OrderBy(c => Rand.Next()).ToArray();
    var result = new StringBuilder(size);
    for (int i = 0, loopTo = size - 1; i <= loopTo; i++)
      result.Append(alpha[Rand.Next(alpha.Length)]);
    return result.ToString();
  }

  private static string AlphaString()
  {
    var result = new StringBuilder(62);
    var z = new byte[] { 48, 57, 65, 90, 97, 122 };
    for (int i = 0, loopTo = z.Length - 1; i <= loopTo; i += 2)
    {
      for (byte j = z[i], loopTo1 = z[i + 1]; j <= loopTo1; j++)
        result.Append(Convert.ToChar(j));
    }
    return result.ToString();
  }

}



internal partial struct InlineableComparable : IComparable<int>
{

  private int m_value;

  public InlineableComparable(int value)
  {
    m_value = value;
  }

  [MethodImpl(MethodImplOptions.AggressiveInlining)]
  public int CompareTo(int other)
  {
    if (m_value == other) return 0;
    if (m_value < other) return -1;
    return 1;
  }
}